import 'dart:async';
import 'dart:developer';
import 'dart:io';
import 'package:cached_network_image/cached_network_image.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter_image_compress/flutter_image_compress.dart';
import 'package:flutter_video_compress/flutter_video_compress.dart';
import 'package:image_picker/image_picker.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_media/model/nui_media_file.dart';
import 'package:nui_media/ui/dialog/nui_select_media_dialog.dart';
import 'package:permission_handler/permission_handler.dart';
import 'package:photo_manager/photo_manager.dart';
import 'package:path_provider/path_provider.dart' as path_provider;
import 'package:path/path.dart' as path;
import 'package:flutter/material.dart';
import 'package:better_player/better_player.dart';
import 'package:video_player/video_player.dart';
import 'package:video_thumbnail/video_thumbnail.dart';

enum MediaType{
  PNG,
  JPG,
  MP4,
  MP4_BIT
}

enum MediaSelectionOption{
  GALLERY,
  CAMERA,
  BOTH
}

enum NUIMediaGalleryOption{
  IMAGE,
  VIDEO,
  BOTH
}
extension NUIMediaGalleryOptionExt on NUIMediaGalleryOption{
  RequestType requestTypeFromOption(){
    if(this == NUIMediaGalleryOption.IMAGE) return RequestType.image;
    if(this == NUIMediaGalleryOption.VIDEO) return RequestType.video;
    return RequestType.all;
  }
}

typedef MediaSelectionListener = void Function(List<NUIMediaFile> medias);

class NUIVideoInfo{
  NUIVideoInfo({this.size, this.width, this.height, this.duration});
  int size;
  double width;
  double height;
  Duration duration;
}

class NUIImageInfo{
  NUIImageInfo({this.size, this.width, this.height});
  int size;
  double width;
  double height;
}

List<NUIMediaFile> showMediaSelection(
    BuildContext context, {
    String title,
    TextStyle titleStyle,
    TextStyle selectionStyle,
    Color iconColor,
    Widget galleryIcon,
    Widget cameraIcon,
    Widget recordIcon,
    bool enableGallery: true,
    bool enableCamera: true,
    bool enableRecord: true}){
  NUIMediaUtil.showMediaSelection(context, title: title, titleStyle: titleStyle, selectionStyle: selectionStyle, iconColor: iconColor, galleryIcon: galleryIcon, cameraIcon: cameraIcon, recordIcon: recordIcon, enableGallery: enableGallery, enableCamera: enableCamera, enableRecord: enableRecord);
}

class NUIMediaUtil{
  NUIMediaUtil._builder();

  static double heightByMediaRatio(BuildContext context, NUIMediaFile file, double width){
    final ratio = file.processHeightToWidthRatio();
    return width * ratio;
  }

  static Future<NUIMediaFile> takePhoto(BuildContext context, {double maxWidth: 1080, double maxHeight: 1080, int imageQuality: 90, bool selfie: false}) async{
    final file = await ImagePicker().getImage(source: ImageSource.camera, maxWidth: maxWidth, maxHeight: maxHeight, imageQuality: imageQuality, preferredCameraDevice: selfie == true ? CameraDevice.front : CameraDevice.rear);
    final imageInfo = await getImageInfo(file.path);
    return NUIMediaFile(path: file.path, width: imageInfo.width.toInt(), height: imageInfo.height.toInt(), isLocal: true);
  }

  static Future<NUIMediaFile> recordVideo(BuildContext context, {double maxWidth: 1080, double maxHeight: 1080, int imageQuality: 90, bool selfie: false}) async{
    final file = await ImagePicker().getVideo(source: ImageSource.camera, preferredCameraDevice: selfie == true ? CameraDevice.front : CameraDevice.rear);
    final videoInfoFut = getVideoInfo(file.path);
    final videoThumbFut = VideoThumbnail.thumbnailData(
      video: file.path,
      imageFormat: ImageFormat.JPEG,
      maxWidth: maxWidth.toInt(), // specify the width of the thumbnail, let the height auto-scaled to keep the source aspect ratio
      maxHeight: maxHeight.toInt(),
      quality: imageQuality ?? 90,
    );

    final videoInfo = await videoInfoFut;
    final mediaFile = NUIMediaFile(path: file.path, videoPath: file.path, thumbData: videoThumbFut, width: videoInfo.width.toInt(), height: videoInfo.height.toInt(), isLocal: true, type: NUIMediaFile.MEDIA_TYPE_VIDEO);
    mediaFile.videoDuration = videoInfo.duration.inMilliseconds;
    return mediaFile;
  }

  static double widthByMediaRatio(BuildContext context, NUIMediaFile file, double height){
    final ratio = file.processHeightToWidthRatio();
    return height / ratio;
  }

  static Future<bool> requestStoragePermissions() async{
    return NUIPermissionUtil.requestPermissions([Permission.storage]);
  }

  static Future<List<AssetEntity>> getAllFromGallery({RequestType type: RequestType.all, int limit:100}) async {
    await requestStoragePermissions();
    var result = await PhotoManager.requestPermission();
    if(result){
      List<AssetEntity> finalImageList = List();
      List<AssetPathEntity> list = await PhotoManager.getAssetPathList(type: type);
      int galleryCount = 0;
      for(AssetPathEntity ape in list){
        final intRemainingSlot = limit - galleryCount;
        if(intRemainingSlot > 0) {
          final imageList = await ape.getAssetListRange(start: galleryCount, end: galleryCount + intRemainingSlot);
          finalImageList.addAll(imageList);
          galleryCount = finalImageList.length;
        }
      }
      return finalImageList;
    }
    else{
      return null;
    }
  }

  static Future<List<NUIMediaFile>> getAllFromGalleryAsNUIMedias({NUIMediaGalleryOption option: NUIMediaGalleryOption.BOTH, int limit:100}) async {
    await requestStoragePermissions();
    var result = await PhotoManager.requestPermission();
    logNUI("NUIMediaUtil", "Requested permission: $result");
    if(result){
      List<NUIMediaFile> finalImageList = [];
      List<AssetPathEntity> list = await PhotoManager.getAssetPathList(type: option.requestTypeFromOption());
      int galleryCount = 0;
      for(AssetPathEntity ape in list){
        final intRemainingSlot = limit - galleryCount;
        if(intRemainingSlot > 0) {
          final imageList = await ape.getAssetListRange(start: galleryCount, end: galleryCount + intRemainingSlot);
          imageList.forEach((element) async{
            if(element.type == AssetType.image || element.type == AssetType.video) {
              final path = (await element.file).path;
              bool shouldRotate = false;
              if(element.orientation == 90 || element.orientation == 270){
                shouldRotate = true;
              }
              finalImageList.add(NUIMediaFile(
                  path: path,
                  isLocal: true,
                  width: shouldRotate ? element.height : element.width,
                  height: shouldRotate ? element.width : element.height,
                  createdDate: element.modifiedDateTime,
                  thumbData: element.type == AssetType.video
                      ? element.thumbData
                      : null,
                  videoPath: element.type == AssetType.video ? path : null,
                  type: element.type == AssetType.video ? NUIMediaFile.MEDIA_TYPE_VIDEO : NUIMediaFile.MEDIA_TYPE_IMAGE
                ),
              );
            }
          });
          galleryCount = finalImageList.length;
        }
      }

      finalImageList.sort((first, second){
        return second.createdDate.compareTo(first.createdDate);
      });
      return finalImageList;
    }
    else{
      return null;
    }
  }

  static Future<AssetEntity> getMediaById(String id) async{
    return await AssetEntity.fromId(id);
  }

  static Size getCompressRatioSize({double maxWidth: 720, double maxHeight: 720, @required double actualWidth, @required double actualHeight}){
    final ratio = actualHeight/actualWidth;
    final isPortrait = actualHeight > actualWidth;

    double newWidth;
    double newHeight;
    if(isPortrait){
      newHeight = maxHeight;
      newWidth = newHeight/ratio;
    }
    else{
      newWidth = maxWidth;
      newHeight = newWidth * ratio;
    }
    return Size(newWidth, newHeight);
  }

  static Future<NUIMediaFile> compressImage(File file, {String targetPath, double maxWidth:1080, double maxHeight: 1080, int quality: 100, bool webpFormat: false}) async {
    final imageInfo = await NUIMediaUtil.getImageInfo(file.path);
    final targetSize = getCompressRatioSize(actualWidth: imageInfo.width, actualHeight: imageInfo.height, maxHeight: maxHeight, maxWidth: maxWidth);

    logNUI("NUMedialUtil", "Compressing image source : ${await file.length()}, path: ${file.path}");

    final tempFile = await getOutputMediaFile(MediaType.JPG, createFirst: true, extension: path.extension(file.path));
    final inSampleSize = calculateInSampleSize(oriWidth: imageInfo.width.toInt(), oriHeight: imageInfo.height.toInt(), targetHeight: targetSize.height.toInt(), targetWidth: targetSize.width.toInt());

    logNUI("MUIMediaUtil", "In sample size: $inSampleSize, oriHeight: ${imageInfo.height}, oriWidth: ${imageInfo.width}, targetWidth: ${targetSize.width}, targetHeight: ${targetSize.height}");

    var result = await FlutterImageCompress.compressAndGetFile(
      file.absolute.path, targetPath ?? (tempFile.path),
      quality: quality ?? 100,
      autoCorrectionAngle: true,
      minHeight: targetSize.height.toInt(),
      minWidth: targetSize.width.toInt(),
//      format: CompressFormat.webp,
      inSampleSize: inSampleSize
    );

    logNUI("NUIMediaUtil", "Compressed image result : ${await result.length()}, path : ${result.path}");
    return NUIMediaFile(path: result.path, width: targetSize.width.toInt(), height: targetSize.height.toInt(), isLocal: true);
  }

  static int calculateInSampleSize({int oriWidth, int oriHeight, int targetWidth, int targetHeight}) {
    var inSampleSize = 1;
    if (oriHeight >= targetHeight || oriWidth >= targetWidth) {
      while (oriWidth/ inSampleSize / 2 >= targetWidth && oriHeight / inSampleSize / 2 >= targetHeight) {
        inSampleSize *= 2;
      }
    }
    return inSampleSize;
  }


  static Future<NUIMediaFile> compressVideoWithBitmap(File file, {String targetPath, int quality, VideoQuality videoQuality: VideoQuality.MediumQuality, bool deleteOrigin: false, bool includeAudio: true, double maxHeight: 720, double maxWidth: 720}) async{
    final imageInfo = await NUIMediaUtil.getVideoInfo(file.path);
    final targetSize = getCompressRatioSize(actualWidth: imageInfo.width, actualHeight: imageInfo.height, maxHeight: maxHeight, maxWidth: maxWidth);

    final mediaFile = NUIMediaFile(path: null, width: targetSize.width.toInt(), height: targetSize.height.toInt());
    final bitmap = bitmapFromVideo(file.path, ImageFormat.PNG, quality: quality, maxHeight: maxHeight, maxWidth: maxWidth);
    final tempFile = await getOutputMediaFile(MediaType.JPG, createFirst: true, extension: path.extension(file.path));

    final video = compressVideo(file, targetPath: targetPath ?? tempFile.path, quality: videoQuality, deleteOrigin: deleteOrigin, includeAudio: includeAudio);
    final bitmapResult = await bitmap;
    final videoResult = await video;
    mediaFile.path = bitmapResult.path;
    mediaFile.videoPath = videoResult.path;
    mediaFile.type = NUIMediaFile.MEDIA_TYPE_VIDEO;
    mediaFile.isLocal = true;
    return mediaFile;
  }

  static Future<MediaInfo> compressVideo(File file, {String targetPath, VideoQuality quality: VideoQuality.MediumQuality, bool deleteOrigin: false, bool includeAudio: true}) async{
    final vdCompress = FlutterVideoCompress();
    logNUI("NUMediaUtil", "Compressing video source : ${await file.length()}, path: ${file.path}");
    final media = await vdCompress.compressVideo(file.path, deleteOrigin: deleteOrigin, includeAudio: includeAudio, quality: quality ?? VideoQuality.MediumQuality);
    logNUI("NUMediaUtil", "Compressed video result : ${await media.file.length()}, path: ${media.file.path}");
    return media;
  }
  
  static Future<File> copyTempFile(File file, {String appendix: "temp_", bool withExtension: true}) async{
    return getFile(nameFromPath(file.path, withExtension: withExtension), appendix: appendix);
  }

  static String fileNameFromTimestamp(String filePath){
    final extension = path.extension(filePath);
    final fileName = path.basenameWithoutExtension(filePath);
    final timestamp = dateToString(DateTime.now(), format: NUIDTUtil.FORMAT_yyyyMMdd_HHmmss).replaceAll("-", "").replaceAll(":", "").replaceAll(" ", "_");
    final uuid = randomUUID().replaceAll("-", "");
    final newFileName = "NUI_${timestamp}_${uuid}${extension}";
    return newFileName;
  }

  static String nameFromPath(String filePath, {bool withExtension: true}){
    if(withExtension == true) {
      return path.basename(filePath);
    }
    return path.basenameWithoutExtension(filePath);
  }

  static Future<List<NUIMediaFile>> compressMediaFiles(List<NUIMediaFile> files) async{
    final hasPermission = await NUIPermissionUtil.requestPermission(Permission.storage);
    if(!hasPermission){
      logNUI("NUIMediaUtil", "You have to grant the storage permission first");
      return null;
    }

    final futureList = List<Future<NUIMediaFile>>();
    for(NUIMediaFile file in files){
      if(file.isLocal != true) continue;

      if(file.isImage()){
        logNUI("NUIMediaUtil", "Compressing image from path : ${file.path}");
        futureList.add(compressImage(file.getFile()));
      }
      else{
        logNUI("NUIMediaUtil", "Compressing video from path : ${file.path}");
        futureList.add(compressVideoWithBitmap(file.getFile()));
      }
    }
    return Future.wait(futureList);
  }

  static Future<File> bitmapFromVideo(String path, ImageFormat format, {double maxWidth: 720, double maxHeight: 720, quality: 90}) async{
    logNUI("NUMediaUtil", "Compressing video thumbnail for video path : $path");
    final thumbnailFile = await FlutterVideoCompress().getThumbnailWithFile(
        path,
        quality: quality ?? 90, // default(100)
        position: -1 // default(-1)
    );
    logNUI("NUMediaUtil", "Compressed video thumbnail path: ${thumbnailFile.path} successful");
    return thumbnailFile;
  }
  
  static Future<File> getOutputMediaFile(MediaType type, {bool createFirst: true, String extension}) async{
    // Create activity media file name
    final uuid = randomUUID().replaceAll("-","");
    final timeStamp = NUIDTUtil.dateToString(DateTime.now(), format: NUIDTUtil.FORMAT_yyyyMMdd_HHmmss) + "_" + uuid;

    if(!isNullOrEmpty(extension)){
      return getFile("NUI_"+ timeStamp + extension);
    }

    String filename;
    if (type == MediaType.JPG) {
      filename = ("IMG_" + timeStamp + ".jpg");
    }
    else if (type == MediaType.MP4) {
      filename = ("VID_" + timeStamp + ".mp4");
    }
    else if (type == MediaType.MP4_BIT){
      filename = ("VID_" + timeStamp + "_video.jpg");
    }
    else if(type == MediaType.PNG){
      filename = ("IMG_" + timeStamp + ".png");
    }
    else{
      filename = ("IMG_" + timeStamp + ".jpg");
    }

    if(createFirst){
      return getFile(filename);
    }
    else{
      return File(await getFilePath(filename));
    }
    //return mediaFile;
  }

  static Future<Directory> getTemporaryDirectory() async {
    return Directory.systemTemp;
  }

  static Future<String> getFilePath(String filename) async {
    final dir = await path_provider.getTemporaryDirectory();
    return "${dir.absolute.path}/$filename";
  }

  static Future<File> getFile(String filename, {String appendix: "temp_"}) async {
    final dir = await path_provider.getTemporaryDirectory();
    File file = createFile("${dir.absolute.path}/$appendix$filename");
    return file;
  }

  static File createFile(String path) {
    final file = File(path);
    if (!file.existsSync()) {
      file.createSync(recursive: true);
    }
    return file;
  }

  static Future<NUIImageInfo> getImageInfo(String imagePath, {bool local: true}) async {
    Completer<NUIImageInfo> completer = Completer();
    Image image = local ? Image.file(File(imagePath)) : Image.network(imagePath);
    image.image.resolve(ImageConfiguration()).addListener(
      ImageStreamListener((ImageInfo image, bool synchronousCall) async{
          var myImage = image.image;
          final fileSize = local ? await File(imagePath).length() : null;
          Size size = Size(myImage.width.toDouble(), myImage.height.toDouble());
          completer.complete(NUIImageInfo(size: fileSize, width: size.width, height: size.height));
        },
      ),
    );
    return completer.future;
  }

  static Future<NUIVideoInfo> getVideoInfo(String videoPath) async{
    final file = File(videoPath);
    final controller = VideoPlayerController.file(file);
    await controller.initialize();
    final duration = controller.value.duration;
    final width = controller.value.size.width;
    final height = controller.value.size.height;
    final size = await file.length();
    return NUIVideoInfo(size: size, duration: duration, width: width, height: height);
  }

  static Future<NUIMediaFile> fileToMediaFile(File file, {bool local: true, MediaType type}) async{
    switch(type){
      case MediaType.JPG:
      case MediaType.MP4_BIT:
      case MediaType.PNG: {
        final imageInfo = await getImageInfo(file.path, local: local);
        return NUIMediaFile(path: file.path, width: imageInfo.width.toInt(), height: imageInfo.height.toInt());
      }
      case MediaType.MP4: {
        final videoInfo = await getVideoInfo(file.path);
        final mediaFile = NUIMediaFile(path: file.path, width: videoInfo.width.toInt(), height: videoInfo.height.toInt());
        mediaFile.videoDuration = videoInfo.duration.inMilliseconds;
        return mediaFile;
      }
      default: {
        final imageInfo = await getImageInfo(file.path, local: local);
        return NUIMediaFile(path: file.path, width: imageInfo.width.toInt(), height: imageInfo.height.toInt());
      }
    }
  }

  static Future<NUIMediaFile> chooseImageFromGallery({ImageSource source: ImageSource.gallery, int imageQuality : 100, double maxWidth, double maxHeight}) async{
    final ImagePicker imagePicker = ImagePicker();
    PickedFile image = await  imagePicker.getImage(
        source: source ?? ImageSource.gallery,
        imageQuality: imageQuality ?? 100,
        maxWidth: maxWidth,
        maxHeight: maxHeight
    );
    if(image != null) {
      return fileToMediaFile(File(image.path), local: true, type: MediaType.PNG);
    }
    return null;
  }

  static Future<NUIMediaFile> chooseVideoFromGallery({ImageSource source: ImageSource.gallery, Duration maxDuration}) async{
    final ImagePicker imagePicker = ImagePicker();
    PickedFile image = await  imagePicker.getVideo(
        source: source ?? ImageSource.gallery,
        maxDuration: maxDuration
    );
    return fileToMediaFile(File(image.path), local: true, type: MediaType.MP4);
  }

  static NUIBottomSheet showMediaSelection(
      BuildContext context, {
        String title,
        TextStyle titleStyle,
        TextStyle selectionStyle,
        Color iconColor,
        Widget galleryIcon,
        Widget cameraIcon,
        Widget recordIcon,
        bool enableGallery: true,
        bool enableCamera: true,
        bool enableRecord: true,
        MediaSelectionListener listener,
      }){
    final dialog = showNUIBottomSheet(context: context, dismissible: true, content: (context) => NUISelectMediaDialogUI(title: title, titleStyle: titleStyle, selectionStyle: selectionStyle, iconColor: iconColor, galleryIcon: galleryIcon, cameraIcon: cameraIcon, recordIcon: recordIcon, enableCamera: enableCamera, enableRecord: enableRecord, enableGallery: enableGallery, listener: listener));
    return dialog;
  }

  static Widget loadImage(String imageUrl, double width, double height, {BoxFit fit: BoxFit.cover, Widget placeHolder, Widget errorWidget}){
    final targetPlaceHolder = placeHolder != null ? placeHolder : Center(child: SizedBox(height: 30.0, width: 30.0, child: CircularProgressIndicator(strokeWidth: 2.0)));
    return CachedNetworkImage(
      imageUrl: imageUrl,
      width: width,
      height: height,
      fit: fit,
      placeholder: (context, url) => targetPlaceHolder,
      errorWidget: (context, url, error) => errorWidget,
    );
  }

  static Widget loadMediaImage(NUIMediaFile mediaFile, double width, double height, {BoxFit fit: BoxFit.cover, Widget placeHolder, Widget errorWidget}){
    final targetPlaceHolder = placeHolder != null ? placeHolder : Center(child: SizedBox(height: 30.0, width: 30.0, child: CircularProgressIndicator(strokeWidth: 2.0)));
    return CachedNetworkImage(
      imageUrl: mediaFile.path,
      width: width,
      height: height,
      fit: fit,
      placeholder: (context, url) => targetPlaceHolder,
      errorWidget: (context, url, error) => errorWidget,
    );
  }

  static ImageProvider loadImageProvider(String imageUrl){
    return CachedNetworkImageProvider(imageUrl);
  }

  static Widget loadImageLocal(String path, double width, double height, {BoxFit fit:BoxFit.cover}){
    return Image.file(File(path), width: width, height: height, fit: fit,);
  }

  static BetterPlayerController getVideoController(NUIMediaFile mediaFile, {BetterPlayerConfiguration config: const BetterPlayerConfiguration()}){
    if(mediaFile.isLocal){
      BetterPlayerDataSource betterPlayerDataSource = BetterPlayerDataSource(BetterPlayerDataSourceType.file, mediaFile.videoPath);
       return BetterPlayerController(BetterPlayerConfiguration(), betterPlayerDataSource: betterPlayerDataSource);
    }
    else{
      BetterPlayerDataSource betterPlayerDataSource = dataSourceFromNUIMedia(mediaFile);
      return BetterPlayerController(config, betterPlayerDataSource: betterPlayerDataSource);
    }
  }

  static BetterPlayerDataSource dataSourceFromNUIMedia(NUIMediaFile media, {bool useCache: true}){
    return BetterPlayerDataSource(
        media.isLocal ? BetterPlayerDataSourceType.file : BetterPlayerDataSourceType.network,
        media.videoPath,
        cacheConfiguration: BetterPlayerCacheConfiguration(useCache: useCache ?? true));
  }

  static List<BetterPlayerDataSource> dataSourceFromNUIMedias(List<NUIMediaFile> mediaFiles){
    List dataSourceList = List<BetterPlayerDataSource>();
    for(NUIMediaFile media in mediaFiles){
      dataSourceList.add(
        dataSourceFromNUIMedia(media)
      );
    }
    return dataSourceList;
  }

}